'' =================================================================================================
''
''   File....... hc-8+_vixen_ssc_player_v3.spin
''   Purpose.... Plays Vixen show from uSD to SSC
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2012 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 16 AUG 2012
''
'' =================================================================================================

{{

  Plays binary file exported from Vixen

  For playback of exported files the EFX-TEK microSD adapter must be connected to the option
  bins located just above the Options switch.  All Option switches must be moved to the OFF
  postion for microSD card access.

  -- triggered by high on IN0
  -- starts audio file on AP-16+ (uses SERIAL header) 
  -- servo output is to Mini-SSC servo controllers (uses OUT0 TTL header)

}}


con

  PCB_REV = "C"


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  

con

  RX1     = 31                                                  ' programming / debug port
  TX1     = 30
  
  SDA     = 29                                                  ' eeprom
  SCL     = 28

  OUT_EN  = 27                                                  ' outputs enable pin (low)     
  
  SIO     = 26                                                  ' TTL serial io
  
  RX2     = 23                                                  ' RS-485 port
  RX2_EN  = 22
  TX2_EN  = 21
  TX2     = 20

  R_LED   = 18                                                  ' bi-color LED
  G_LED   = 17

  OUT7    = 15                                                  ' outputs
  OUT6    = 14   
  OUT5    = 13 
  OUT4    = 12
  OUT3    = 11
  OUT2    = 10   
  OUT1    =  9 
  OUT0    =  8
          
  OPT_BR  =  7                                                  ' options swithes
  OPT_A1  =  6   
  OPT_A0  =  5
  OPT_SM  =  4

  DMX_A8  =  3                                                  ' high-bit off DMX address

  DO_165  =  2                                                  ' control lines for 74x165
  CLK_165 =  1 
  LD_165  =  0


  ' for EFX-TEK uSD adapter module (#99002)
  ' -- must open all option switches when installed

  SD_CS   =  OPT_BR                                       
  SD_DI   =  OPT_A1  
  SD_CLK  =  OPT_A0  
  SD_DO   =  OPT_SM


con

   #0, NO, YES

   #0, OFF, RED, GRN, YEL                                       ' for r/g led

   #0, NOCON, RX, TX, RXTX                                      ' for rs-485

   #0, LSBFIRST, MSBFIRST                                       ' for 74x165 shift-register                                               


con

   #1, HOME, GOTOXY, #8, BKSP, TAB, LF, CLREOL, CLRDN, CR       ' PST formmatting control
  #14, GOTOX, GOTOY, CLS


con

  SBAUD     =  9_600                                            ' match Mini-SSC baud
  ABAUD     = 38_400                                            ' fast EFX-TEK baud

  CH_COUNT  = 16                                                ' channels in data file
  CH_MASK   = %00011101_00111111                                ' active channels


obj

  sd     : "fsrw"                                               ' SD card IO
  servos : "fullduplexserial64"                                 ' serial input (from Vixen)
  audio  : "fullduplexserial64"                                 ' for serial commands to AP-16+
      

var

  long  mounted                                                 ' true when SD mounted

  word  channels                                                ' channels in Vixen project
  word  eventms                                                 ' event timing; milliseconds
  word  frames                                                  ' frames (events) in project
  
  byte  buffer[CH_COUNT]                                        ' frame buffer
  

dat

ShowFile                byte    "vixen.bin", 0                  ' show file name (on uSD)
AudioFile               byte    "triller.wav", 0                ' audio file name (on AP-16+)

  
pub main | check

  setup_io                                                      ' configure HC-8 io pins
  servos.start(OUT0, OUT0, %1011, SBAUD)                        ' * inverted serial for Mini-SSCs
  audio.start(SIO, SIO, %1100, ABAUD)                           ' true serial for AP-16+

  repeat                                                        ' attempt to mount SD card
    check := mount_sd
    if (check)
      set_rgled(GRN)
      quit
    else
      set_rgled(RED)
 
  pause(3500)                                                   ' let ap16 power up

  repeat
    stop_ap16                                                   ' stop if playing
        
    repeat                                                      ' wait for IN0 high 
      check := ttl_pin
    until (check == 0)                                          

    play_ap16(@AudioFile)

    { may need to add time pad here }    

    play_vixen(@ShowFile)


pub play_ap16(spntr)

'' Plays wav file defined by spntr

  audio.str(string("!AP16"))                                    ' header
  audio.tx(%00)                                                 ' address
  audio.str(string("PW"))                                       ' play WAV
  audio.str(spntr)                                              ' filename
  audio.tx(CR)                                                  ' required CR
  audio.tx(1)                                                   ' play one time


pub stop_ap16

'' Stops/resets AP-16+

  audio.str(string("!AP16"))                                    ' header
  audio.tx(%00)                                                 ' address
  audio.tx("X")                                                 ' reset command


pub play_vixen(fpntr) | t, ch

'' Plays data in file defined by fpntr

  if (open_file(fpntr, "r"))                                    ' open show file
    if (readbuf(@channels, 6))                                  ' read header (3 words)
      set_rgled(GRN)
      t := cnt
      repeat
        if (readbuf(@buffer, channels))                         ' read channels for one frame
          repeat ch from 0 to CH_COUNT-1
            if (CH_MASK & (1 << ch))                            ' active channel?
              servos.tx($FF)                                    '  sync byte
              servos.tx(ch)                                     '  channel
              servos.tx(buffer[ch])                             '  position

          toggle(G_LED)                                         ' show running    
          waitcnt(t += (MS_001 * eventms))                      ' hold for frame timing                    
        else
          close_file
          set_rgled(GRN)                                        ' leave on
          quit

    else
      set_rgled(RED)

  else
    set_rgled(RED)
        
         

con

  { ================= }
  {                   }
  {  F I L E   I / 0  } 
  {                   }
  { ================= }


pub mount_sd | check

'' Attempt to mount SD card
'' -- returns true if SD card successfully mounted

  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)

  if (check => 0)
    mounted := true
  else
    mounted := false

  return mounted


pub dismount_sd

'' Unmounts SD card

  \sd.unmount
  mounted := false


pub open_file(spntr, mode) | check

'' Attempt to open file
'' -- spntr is pointer to filename (z-string)
'' -- mode is single character, (e.g., "r" for read)
'' -- sd card must be mounted before open

  if (mounted)
    check := \sd.popen(spntr, mode)                             ' attempt to open
    if (check => 0)
      if ((mode == "r") or (mode == "R"))
        \sd.seek(0)                                             ' force sector read 
      return true
    else
      return false
  else
    return false


pub close_file

'' Closes open file

  \sd.pclose


pub readln(spntr, n) | c, len

'' Reads line of text from open text file
'' -- terminated by LF or EOF
'' -- spntr is address of string buffer
'' -- n is maximum number of characters to read from line

  len := 0                                                     ' index into string

  repeat
    c := \sd.pgetc                                              ' get a character
    if (c < 0)                                                  ' end of file
      if (len == 0)                                             ' if nothing
        return -1                                               ' return empty
      else
        quit

    if (c <> LF)                                                ' if not EOL char
      byte[spntr++] := c                                        '  add to buffer
      len++
    else                                                        ' if LF
      quit                                                      '  we're done

  byte[spntr] := 0                                              ' terminate end of line

  return len
  

pub writeln(spntr, usecr, uself)

'' Writes z-string to open text file
'' -- set usecr to true to add CR
'' -- set uself to true to add LF

  \sd.pputs(spntr)
  if (usecr)
    \sd.pputc(CR)
  if (uself)
    \sd.pputc(LF)


pub readbuf(bpntr, n) | check

'' Reads n bytes from file to buffer at bpntr
'' -- returns true if good read

  check := \sd.pread(bpntr, n)
  if (check == n)
    return true
  else
    return false 


pub writebuf(bpntr, n) | check   

'' Writes n bytes to file from buffer at bpntr
'' -- returns true if good read

  check := \sd.pwrite(bpntr, n)
  if (check == n)
    return true
  else
    return false

      
con

  { ================================= }
  {                                   }
  {  S U P P O R T   R O U T I N E S  } 
  {                                   }
  { ================================= }


var

  long  lastscan                                                ' ttl/DMX input scan


pub setup_io

'' Configure for basic IO control

  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode                                  
  low(OUT_EN)                                                   ' enable 74x245 

  high(LD_165)                                                  ' setup x165 io pins
  low(CLK_165)
  input(DO_165)
  
  set_rs485(NOCON)                                              ' RS-485 off


pub pause(ms) | t

'' Delay program in milliseconds

  if (ms < 1)                                                   ' delay must be at least 1
    return
  else
    t := cnt - 1792                                             ' sync with system counter
    repeat ms                                                   ' run delay
      waitcnt(t += MS_001)


pub high(pin)

'' Makes pin output and high

  outa[pin] := 1
  dira[pin] := 1


pub low(pin)

'' Makes pin output and low

  outa[pin] := 0
  dira[pin] := 1


pub toggle(pin)

'' Toggles output pin

  !outa[pin]                                                    ' invert state
  dira[pin] := 1  
  

pub input(pin)

  dira[pin] := 0

  return ina[pin]


pub ttl_pin | btns

'' Reads TTL header for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  btns := ttl_inputs(true)                                      ' read ttl inputs

  if (bit_count(btns) == 1)                                     ' only one button pressed?
    return bit_pos(btns, 0)                                     '  yes, return button #
  else
    return -1                                                   '  none or bad input


pub ttl_inputs(rescan) 

'' Return TTL inputs status
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss
   
  return (lastscan & $FF)                                       ' return TTL input bits


pub dmx_address(rescan) 

'' Return DMX address switch setting
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss

  return (lastscan >> 8) & $01FF                                ' return DMX address bits
 

pub x165_in(bits, mode) | tmp165

'' Input value from 74x165(s)      
                                                                                          
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                                                         
                                                                                          
  tmp165 := 0                                                   ' clear workspace
  bits := 1 #> bits <# 32                                       ' limit to legal value         
  repeat bits                                                   ' read n bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                       ' get new bit             
    outa[CLK_165] := 1                                          ' blip clock              
    outa[CLK_165] := 0  

  if (mode == LSBFIRST)                                         ' LSBFIRST result?
    tmp165 ><= bits                                             ' reverse bits
   
  return tmp165
  

pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                 ' force to input mode

  return !ina[OPT_BR..OPT_SM]                                   ' return state (1 = on)


pub set_rgled(state)

'' Sets R/G LED
'' -- direct control only
'' -- for red/green/yellow use "efx_rgled" driver and methods

  case state
    OFF : outa[R_LED..G_LED] := %00
    RED : outa[R_LED..G_LED] := %10
    GRN : outa[R_LED..G_LED] := %01
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires PASM driver

  dira[R_LED..G_LED] := %11
      

pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(target) | bc

'' Returns # of bits set (1) in target

  bc := 0                                                       ' clear count
  repeat 32                                                     ' test all bits
    bc += (target & %1)                                         ' add bit value to count
    target >>= 1                                                ' next bit

  return bc


pub bit_pos(target, mode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb 
'' -- -1 = no bits set

  if (target == 0)                                              ' if no bits                                   
    return -1                                                   '  return -1

  else
    if (mode == 0)                                              ' check from LSB
      target ><= 32                                             '  flip for >|
      return (32 - >|target)
    else
      return (>|target - 1)


pub bit_val(target, pos)

'' Returns bit (0..1) from pos (0..31) in target 

  if ((pos => 0) and (pos =< 31))                               ' if good bit position
    return (target >> pos) & 1                                  '  extract bit
  else
    return 0


pub set_bit(target, pos, newbit)

  if ((pos => 0) and (pos =< 31))                               ' if good bit position
    if (newbit)                                                 '  if true/1
      target |= (1 << pos)                                      '   set the bit                                     
    else                                                        '  if false/0
      target &= !(1 << pos)                                     '   clear the bit

  return target  


dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      